import re
import spacy

#build spacy model for procees text
spacy.require_gpu()
en_nlp=spacy.load("en_core_web_md")


#replace html element in context
NonHtml=re.compile(r"<[^<]+?>|<!--.*?-->|&([a-z0-9]+|#[0-9]{1,6}|#x[0-9a-f]{1,6});")

def strip_html(text):
    text=NonHtml.sub("",text)
    return text


import os
import logging
import copy
import json
import torch
import spacy
import random
from collections import namedtuple
from utils import load_posTag

logger = logging.getLogger(__name__)
InputFeature=namedtuple("Example",['uid','sents','label'])
RnnFeature=namedtuple("RnnExample", ['sents','pos_sents','label'])


class textExtracter:
    def __init__(self, nlp_name,use_gpu,max_sents):
        if use_gpu:
            spacy.require_gpu()
        self.nlp_pipe = spacy.load(nlp_name)
        self.max_sents=max_sents
    @classmethod
    def _read_file(self, folder_path,data_num):
        lines = []
        for i,f_name in os.listdir(folder_path):
            with open(os.path.join(folder_path,f_name), 'r', encoding='utf-8') as f_r:
                for line in f_r:
                    line = line.strip()
                    lines.append(line)
            if i==data_num:
                break
        return lines

    def create_examples(self, set_type,input_text,label):
        examples = []
        for idx,text in enumerate(input_text):
            uid = '%s-%s' % (set_type, idx)
            # split text to sents
            doc=self.nlp_pipe(text.strip())
            sents=[sent.text for sent in doc.sents][:self.max_sents]           
            # add to examples
            examples.append(InputFeature(uid=uid, sents=sents,label=label))
        return examples

    def get_examples(self,data_dir,mode,data_num):
        '''
        Mode:train,test
        '''
        data_path = os.path.join(self.args.data_dir, mode)
        labels={'pos':1,'neg':0}
        #create sents for pos & neg data
        examples=[]
        for l in ['pos','neg']:
            f_path=os.path.join(data_dir,l)
            examples+=self.create_examples(mode,self._read_file(f_path,data_num),
                                           label=labels[l])
        examples=random.sample(examples)
        return examples

def convert_to_RnnFeatures(data, max_seqLen,
                           tokenizer,pos_tags):
    # 整理examples features
    examples = []
    for ex_id, example in enumerate(data):
        if ex_id % 5000 == 0:
            logger.info(f'Already convert {ex_id} examples!')
        #extract linguistic feature for ecah sent in one example
        sents_id=[]
        pos_sents=[]
        for s in data.sents:
            sent=self.nlp_pipe(s)
            p=[]
            s_ids=[]
            for t in sent:
                #using bpe tokenizer
                tok_ids=self.tokenizer.encode(t.text).ids
                pos_ids=len(tok_ids)*[pos_tags[t.pos_]]
                s_ids.extend(tok_ids)
                p.extend(pos_ids)
            #truncate value
            s_ids=s_ids[:max_seqLen]
            p=p[:max_seqLen]
            
            assert len(s_ids) == len(p) == max_seqLen
            #add into sents of doc
        sents_id.append(s_ids)
        pos_sents.append(p)

        if ex_id < 5:
            logger.info('****Display examples****')
            logger.info('Uid:{}'.format(example.guid))
            logger.info('token_ids:{}'.format(sents_id))
            logger.info('pos_ids:{}'.format(pos_sents))
            logger.info('label:{}'.format(example.label))

        examples.append(InputRnnFeatures(sents=sents_id,
                                         pos_sents=pos_sents,
                                         label=example.label))

    return examples


def load_and_cacheExampels(args, tokenizer, mode):
    # build extractor
    text_Etl=textExtracter(args.nlp_model,args.use_gpu,args.max_sent)
    pos_tags=load_posTag(args)
    # build load and save cached file path
    cached_file_path = os.path.join(args.data_dir, mode,
                                    'cached_{}_{}.zip'.format(
                                        mode,args.max_seqLen))

    if os.path.isfile(cached_file_path):
        # load features
        logger.info('loading data file from {}'.format(cached_file_path))
        features = torch.load(cached_file_path)
    else:
        # create examples
        if mode == 'train':
            examples = processer.get_examples(args.data_dir,mode,args.max_data_num)
        elif mode == 'test':
            examples = processer.get_examples(args.data_dir,mode,args.max_data_num)
        else:
            raise NameError('Mode args is not train,val,test!')

        if args.model_type.endswith('rnn'):
            features = convert_to_RnnFeatures(examples, args.max_seqLen, tokenizer, pos_tags)
        # save to cached file path
        torch.save(features, cached_file_path)
        logger.info(f'Save features to {cached_file_path}')


    # transform features into tensor
    dataset = TensorDataset(features)

    return dataset

    


import os
from os import path
BASE_DIR=path.abspath(path.dirname(__file__))
#envirmonent settings
OUTPUT_DIR=os.environ.get("OUTPUT_MODELS_DIR",path.join(BASE_DIR,"outputs"))
DATA_DIR=os.environ.get("DATA_DIR",path.join(BASE_DIR,"Data"))

#training configs
MAX_VOCAB_SIZE=25000
SENTS_MAX_LEN=50
SENTS_NUMS=20
MIN_FREQ=2
SPECIAL_TOKENS=["[UNK]","[PAD]"]
ENTITY_MAPPINGS={"O":0,"GPE_B":1,"GPE_I":2,"LANGUAGE_B":3,"LANGUAGE_I":4,"LAW_B":5,"LAW_I":6,"LOC_B":7,
                  "LOC_I":8,"MONEY_B":9,"MONEY_I":10,"NORP_B":11,"NORP_I":12,"ORDINAL_B":13,"ORDINAL_I":14,
                  "ORG_B":15,"ORG_I":16,"PERCENT_B":17, "PERCENT_I":18,"PERSON_B":19, "PERSON_I":20,"PRODUCT_B":21,
                 "PRODUCT_I":22,"QUANTITY_B":23,"QUANTITY_I":24,"TIME_B":25,"TIME_I":26,"WORK_OF_ART_B":27,
                 "WORK_OF_ART_I":21}


from tokenizers import Tokenizer
from tokenizers.models import BPE
from tokenizers.trainers import BpeTrainer
from tokenizers.pre_tokenizers import Whitespace
from tokenizers.normalizers import Sequence,Lowercase,StripAccents
import logging
import argparse
import os

from imdb_prac.settings import *
from imdb_prac.process.text import strip_html,en_nlp
import pandas as pd
#get module name to log message
logger=logging.getLogger(__name__)

def generate_bpe(data_iter,min_freq,max_size):

    save_path=os.path.join(OUTPUT_DIR,"tokenizers")

    #build init tokenizer
    tokenizer=Tokenizer(BPE(unk_token="[UNK]",continuing_subword_prefix="##",
                            end_of_word_suffix="##"))
    trainer=BpeTrainer(special_tokens=["[PAD]","[UNK]"],min_frequency=min_freq,
                       vocab_size=max_size)

    #insert text clean process
    norm_pipe=Sequence([StripAccents(),Lowercase()])
    tokenizer.normalizer=norm_pipe
    tokenizer.pre_tokenizer=Whitespace()
    
    
    tokenizer.train_from_iterator(data_iter,trainer)

    #save trained tokenizer
    if os.path.isdir(save_path):
        logger.info("The model dir has already exists!")
    else:
        os.mkdir(save_path)
        logger.info(f"Create new folder in {save_path}")
    tokenizer.save(os.path.join(save_path,'tokenizer-imdb.json'))

def fetch_raw_data(folder_path,label,data_num):
    i=0
    for i,f_name in os.listdir(folder_path):
        data=[]
        
        with open(os.path.join(folder_path,f_name),"r",encoding='utf-8') as f_r:
            for line in f_r:
                doc=en_nlp(strip_html(line.strip()))
                #seprate line and add symbol for split
                doc="\n".join(list(doc.sents))
                data.append(doc)
        #add label
        data.append(label)
        if i>data_num:
            break
        yield data
        

def create_csv(src_path,dest_path,data_num):
    logger.info(f"------Start to get data from folder:{src_path}-------")
    datas=list(d for l in ['pos','neg'] for d in fetch_raw_data(src_path,l,data_num))
    df=pd.DataFrame(datas,columns=['text','label'])
    df=df.sample(frac=1)
    df.to_csv(dest_path,index=False,encoding="utf-8")


def main(args):
    if args.subparser_name=="data":
        load_path=os.path.join(args.src,args.type)
        create_csv(load_path,args.dst)

    elif args.subparser_name=="token":
        df=pd.read_csv(args.src)
        data_iter=list(sent for r_v in df["text"] for sent in r_v.split("\n") if sent )
        generate_bpe(data_iter,args.min_freq,args.max_size)

if __name__=="__main__":
    parser=argparse.ArgumentParser()
    subparsers=parser.add_subparsers(title="subcommand",dest="subparser_name",
                                     description="data: used for produce data and tokenizer",
                                    help="addition help")
    parser_data=subparsers.add_parser("data",help="data help")
    parser_data.add_argument("src",type=str,default=os.path.join(BASE_DIR,"aclImdb"),help="")
    parser_data.add_argument("dst",type=str,help="")
    parser_data.add_argument("-n","--num",type=int,help="")
    parser_data.add_argument("--type",type=str,choices=["train","test"],
                             help="convert spec data that used for train or testing.")

    #subcommand for training bpe tokenizer
    parser_token=subparsers.add_parser("token",help="token help")
    parser_token.add_argument("src",type=str)
    parser_token.add_argument('-mf',"--minfreq",type=int,default=1)
    parser_token.add_argument("-ms","--max_size",type=int,default=25000)
    
    args=parser.parse_args()
    main(args)

                 
                






import os
import logging
import json
import random
import torch
import numpy as np
from tokenizers import Tokenizer
MODEL_CLASSES={
    'h_rnn':(Tokenizer,),
}
MODEL_PATH={
    "h_rnn":".\saved_models\\rnn_model\\",
}

def load_posTag(args):
    with open(os.path.join(args.data_dir,'pos_vocab.json'),'r',encoding='utf-8') as f_r:
        pos_tags=json.load(f_r)
    
    return pos_tags

# set random seed
def set_randomSeed(args):
    random.seed(args.random_seed)
    np.random.seed(args.random_seed)
    torch.random.manual_seed(args.random_seed)
    torch.cuda.manual_seed_all(args.random_seed)


# set logger logging message
def init_logger():
    logging.basicConfig(format='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
                        datefmt=r'%m/%d/%Y %H:%M:%S',
                        level=logging.INFO)
